@@ -250,7 +250,8 @@ class cNMF():
 
 
     def prepare(self, counts_fn, components, n_iter = 100, densify=False, tpm_fn=None, seed=None,
-                         beta_loss='frobenius',num_highvar_genes=2000, genes_file=None):
+                        beta_loss='frobenius',num_highvar_genes=2000, genes_file=None,
+                        alpha_usage=0.0, alpha_spectra=0.0, init='random'):
         """
         Load input counts, reduce to high-variance genes, and variance normalize genes.
         Subsequently prepare file for distributing jobs over workers.
@@ -284,6 +285,12 @@ class cNMF():
         
         genes_file : str or None, optional (default=None)
             If provided will load high-variance genes from a list of these genes
+            
+        alpha_usage : float, optional (default=0.0)
+            Regularization parameter for NMF corresponding to alpha_W in scikit-learn
+
+        alpha_spectra : float, optional (default=0.0)
+            Regularization parameter for NMF corresponding to alpha_H in scikit-learn
         """
         
         
@@ -351,28 +358,36 @@ class cNMF():
 
         norm_counts = self.get_norm_counts(input_counts, tpm, num_highvar_genes=num_highvar_genes,
                                                high_variance_genes_filter=highvargenes)
-        
-        
-        if norm_counts.X.dtype != np.float64:
-            norm_counts.X = norm_counts.X.astype(np.float64)
 
         self.save_norm_counts(norm_counts)
-        (replicate_params, run_params) = self.get_nmf_iter_params(ks=components, n_iter=n_iter, random_state_seed=seed, beta_loss=beta_loss)
+        (replicate_params, run_params) = self.get_nmf_iter_params(ks=components, n_iter=n_iter, random_state_seed=seed,
+                                                                  beta_loss=beta_loss, alpha_usage=alpha_usage,
+                                                                  alpha_spectra=alpha_spectra, init=init)
         self.save_nmf_iter_params(replicate_params, run_params)
         
     
-    def combine(self, components=None):
-        run_params = load_df_from_npz(self.paths['nmf_replicate_parameters'])
+    def combine(self, components=None, skip_missing_files=False):
+        """
+        Combine NMF iterations for the same value of K
+        Parameters
+        ----------
+        components : list or None
+            Values of K to combine iterations for. Defaults to all.
+
+        skip_missing_files : boolean
+            If True, ignore iteration files that aren't found rather than crashing. Default: False
+        """
 
         if type(components) is int:
             ks = [components]
         elif components is None:
+            run_params = load_df_from_npz(self.paths['nmf_replicate_parameters'])
             ks = sorted(set(run_params.n_components))
         else:
             ks = components
 
         for k in ks:
-            self.combine_nmf(k)    
+            self.combine_nmf(k, skip_missing_files=skip_missing_files)    
     
     
     
@@ -454,7 +469,9 @@ class cNMF():
         
     def get_nmf_iter_params(self, ks, n_iter = 100,
                                random_state_seed = None,
-                               beta_loss = 'kullback-leibler'):
+                               beta_loss = 'kullback-leibler',
+                               alpha_usage=0.0, alpha_spectra=0.0,
+                               init='random'):
         """
         Create a DataFrame with parameters for NMF iterations.
 
@@ -471,7 +488,12 @@ class cNMF():
 
         random_state_seed : int or None, optional (default=None)
             Seed for sklearn random state.
+            
+        alpha_usage : float, optional (default=0.0)
+            Regularization parameter for NMF corresponding to alpha_W in scikit-learn
 
+        alpha_spectra : float, optional (default=0.0)
+            Regularization parameter for NMF corresponding to alpha_H in scikit-learn
         """
 
         if type(ks) is int:
@@ -491,14 +513,14 @@ class cNMF():
         replicate_params = pd.DataFrame(replicate_params, columns = ['n_components', 'iter', 'nmf_seed'])
 
         _nmf_kwargs = dict(
-                        alpha_W=0.0,
-                        alpha_H='same',
+                        alpha_W=alpha_usage,
+                        alpha_H=alpha_spectra,
                         l1_ratio=0.0,
                         beta_loss=beta_loss,
                         solver='mu',
                         tol=1e-4,
                         max_iter=1000,
-                        init='random'
+                        init=init
                         )
         
         ## Coordinate descent is faster than multiplicative update but only works for frobenius
@@ -583,35 +605,84 @@ class cNMF():
             save_df_to_npz(spectra, self.paths['iter_spectra'] % (p['n_components'], p['iter']))
 
 
-    def combine_nmf(self, k, remove_individual_iterations=False):
+    def combine_nmf(self, k, skip_missing_files=False, remove_individual_iterations=False):
         run_params = load_df_from_npz(self.paths['nmf_replicate_parameters'])
         print('Combining factorizations for k=%d.'%k)
 
-        combined_spectra = None
-        n_iter = sum(run_params.n_components==k)
-
         run_params_subset = run_params[run_params.n_components==k].sort_values('iter')
-        spectra_labels = []
+        combined_spectra = []
 
         for i,p in run_params_subset.iterrows():
-
-            spectra = load_df_from_npz(self.paths['iter_spectra'] % (p['n_components'], p['iter']))
-            if combined_spectra is None:
-                combined_spectra = np.zeros((n_iter, k, spectra.shape[1]))
-            combined_spectra[p['iter'], :, :] = spectra.values
-
-            for t in range(k):
-                spectra_labels.append('iter%d_topic%d'%(p['iter'], t+1))
-
-        combined_spectra = combined_spectra.reshape(-1, combined_spectra.shape[-1])
-        combined_spectra = pd.DataFrame(combined_spectra, columns=spectra.columns, index=spectra_labels)
-
-        save_df_to_npz(combined_spectra, self.paths['merged_spectra']%k)
+            current_file = self.paths['iter_spectra'] % (p['n_components'], p['iter'])
+            if not os.path.exists(current_file):
+                if not skip_missing_files:
+                    print('Missing file: %s, run with skip_missing=True to override' % current_file)
+                    raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), current_file)
+                else:
+                    print('Missing file: %s. Skipping.' % current_file)
+            else:
+                spectra = load_df_from_npz(current_file)
+                spectra.index = ['iter%d_topic%d' % (p['iter'], t+1) for t in range(k)]
+                combined_spectra.append(spectra)
+                
+        if len(combined_spectra)>0:        
+            combined_spectra = pd.concat(combined_spectra, axis=0)
+            save_df_to_npz(combined_spectra, self.paths['merged_spectra']%k)
+        else:
+            print('No spectra found for k=%d' % k)
         return combined_spectra
+    
+    
+    def refit_usage(self, X, spectra):
+        """
+        Takes an input data matrix and a fixed spectra and uses NNLS to find the optimal
+        usage matrix. Generic kwargs for NMF are loaded from self.paths['nmf_run_parameters'].
+        If input data are pandas.DataFrame, returns a DataFrame with row index matching X and
+        columns index matching index of spectra
+
+        Parameters
+        ----------
+        X : pandas.DataFrame or numpy.ndarray, cells X genes
+            Non-negative expression data to fit spectra to
+
+        spectra : pandas.DataFrame or numpy.ndarray, programs X genes
+            Non-negative spectra of expression programs
+        """
+
+        refit_nmf_kwargs = yaml.load(open(self.paths['nmf_run_parameters']), Loader=yaml.FullLoader)
+        if type(spectra) is pd.DataFrame:
+            refit_nmf_kwargs.update(dict(n_components = spectra.shape[0], H = spectra.values, update_H = False))
+        else:
+            refit_nmf_kwargs.update(dict(n_components = spectra.shape[0], H = spectra, update_H = False))
+            
+        _, rf_usages = self._nmf(X, nmf_kwargs=refit_nmf_kwargs)
+        if (type(X) is pd.DataFrame) and (type(spectra) is pd.DataFrame):
+            rf_usages = pd.DataFrame(rf_usages, index=X.index, columns=spectra.index)
+          
+        return(rf_usages)
+    
+    
+    def refit_spectra(self, X, usage):
+        """
+        Takes an input data matrix and a fixed usage matrix and uses NNLS to find the optimal
+        spectra matrix. Generic kwargs for NMF are loaded from self.paths['nmf_run_parameters'].
+        If input data are pandas.DataFrame, returns a DataFrame with row index matching X and
+        columns index matching index of spectra
+
+        Parameters
+        ----------
+        X : pandas.DataFrame or numpy.ndarray, cells X genes
+            Non-negative expression data to fit spectra to
+
+        usage : pandas.DataFrame or numpy.ndarray, cells X programs
+            Non-negative spectra of expression programs
+        """
+        return(self.refit_usage(X.T, usage.T).T)
 
 
     def consensus(self, k, density_threshold=0.5, local_neighborhood_size = 0.30,show_clustering = True,
-                  skip_density_and_return_after_stats = False, close_clustergram_fig=False):
+                  skip_density_and_return_after_stats = False, close_clustergram_fig=False,
+                  refit_usage=True):
         merged_spectra = load_df_from_npz(self.paths['merged_spectra']%k)
         norm_counts = sc.read(self.paths['normalized_counts'])
 
@@ -624,7 +695,6 @@ class cNMF():
         # Rescale topics such to length of 1.
         l2_spectra = (merged_spectra.T/np.sqrt((merged_spectra**2).sum(axis=1))).T
 
-
         if not skip_density_and_return_after_stats:
             # Compute the local density matrix (if not previously cached)
             topics_dist = None
@@ -661,17 +731,19 @@ class cNMF():
         stability = silhouette_score(l2_spectra.values, kmeans_cluster_labels, metric='euclidean')
 
         # Obtain reconstructed count matrix by re-fitting usage and computing dot product: usage.dot(spectra)
-        refit_nmf_kwargs = yaml.load(open(self.paths['nmf_run_parameters']), Loader=yaml.FullLoader)
-        refit_nmf_kwargs.update(dict(
-                                    n_components = k,
-                                    H = median_spectra.values,
-                                    update_H = False
-                                    ))
-        
-        _, rf_usages = self._nmf(norm_counts.X,
-                                          nmf_kwargs=refit_nmf_kwargs)
-        rf_usages = pd.DataFrame(rf_usages, index=norm_counts.obs.index, columns=median_spectra.index)
+        rf_usages = self.refit_usage(norm_counts.X, median_spectra)
+        rf_usages = pd.DataFrame(rf_usages, index=norm_counts.obs.index, columns=median_spectra.index)        
         rf_pred_norm_counts = rf_usages.dot(median_spectra)
+        
+        # Re-order usage by total contribution
+        norm_usages = rf_usages.div(rf_usages.sum(axis=1), axis=0)      
+        reorder = norm_usages.sum(axis=0).sort_values(ascending=False)
+        rf_usages = rf_usages.loc[:, reorder.index]
+        norm_usages = norm_usages.loc[:, reorder.index]
+        median_spectra = median_spectra.loc[reorder.index, :]
+        rf_usages.columns = np.arange(1, rf_usages.shape[1]+1)
+        norm_usages.columns = rf_usages.columns
+        median_spectra.index = rf_usages.columns
 
         # Compute prediction error as a frobenius norm
         if sp.issparse(norm_counts.X):
@@ -686,46 +758,50 @@ class cNMF():
         if skip_density_and_return_after_stats:
             return consensus_stats
         
-        save_df_to_npz(median_spectra, self.paths['consensus_spectra']%(k, density_threshold_repl))
-        save_df_to_npz(rf_usages, self.paths['consensus_usages']%(k, density_threshold_repl))
-        save_df_to_npz(consensus_stats, self.paths['consensus_stats']%(k, density_threshold_repl))
-        save_df_to_text(median_spectra, self.paths['consensus_spectra__txt']%(k, density_threshold_repl))
-        save_df_to_text(rf_usages, self.paths['consensus_usages__txt']%(k, density_threshold_repl))
-
-        # Compute gene-scores for each GEP by regressing usage on Z-scores of TPM
+        # Convert spectra to TPM units, and obtain results for all genes by running last step of NMF
+        # with usages fixed and TPM as the input matrix
         tpm = sc.read(self.paths['tpm'])
         tpm_stats = load_df_from_npz(self.paths['tpm_stats'])
+        spectra_tpm = self.refit_spectra(tpm.X, norm_usages.astype(tpm.X.dtype))
+        spectra_tpm = pd.DataFrame(spectra_tpm, index=rf_usages.columns, columns=tpm.var.index)
+        spectra_tpm = spectra_tpm.div(spectra_tpm.sum(axis=1), axis=0) * 1e6
         
+        # Convert spectra to Z-score units, and obtain results for all genes by running last step of NMF
+        # with usages fixed and Z-scored TPM as the input matrix
         if sp.issparse(tpm.X):
             norm_tpm = (np.array(tpm.X.todense()) - tpm_stats['__mean'].values) / tpm_stats['__std'].values
         else:
             norm_tpm = (tpm.X - tpm_stats['__mean'].values) / tpm_stats['__std'].values
         
-        if norm_tpm.dtype != np.float64:
-            norm_tpm = norm_tpm.astype(np.float64)
-        
         usage_coef = fast_ols_all_cols(rf_usages.values, norm_tpm)
         usage_coef = pd.DataFrame(usage_coef, index=rf_usages.columns, columns=tpm.var.index)
-
-        save_df_to_npz(usage_coef, self.paths['gene_spectra_score']%(k, density_threshold_repl))
-        save_df_to_text(usage_coef, self.paths['gene_spectra_score__txt']%(k, density_threshold_repl))
-
-        # Convert spectra to TPM units, and obtain results for all genes by running last step of NMF
-        # with usages fixed and TPM as the input matrix
-        norm_usages = rf_usages.div(rf_usages.sum(axis=1), axis=0)
-        refit_nmf_kwargs.update(dict(
-                                    H = norm_usages.T.values,
-                                ))
         
-        # Needed otherwise _nmf will crash because with inconsistent dtypes
-        if tpm.X.dtype != np.float64:
-            tpm.X = tpm.X.astype(np.float64)
-        
-        _, spectra_tpm = self._nmf(tpm.X.T, nmf_kwargs=refit_nmf_kwargs)
-        spectra_tpm = pd.DataFrame(spectra_tpm.T, index=rf_usages.columns, columns=tpm.var.index)
+        if refit_usage:
+            ## Re-fitting usage a final time on std-scaled HVG TPM seems to
+            ## increase accuracy on simulated data
+            hvgs = open(self.paths['nmf_genes_list']).read().split('\n')
+            norm_tpm = tpm[:, hvgs]
+            if sp.issparse(norm_tpm.X):
+                sc.pp.scale(norm_tpm, zero_center=False)                       
+            else:
+                norm_tpm.X /= norm_tpm.X.std(axis=0, ddof=1)
+                
+            spectra_tpm_rf = spectra_tpm.loc[:,hvgs]
+            tpm_stats.index = tpm.var.index
+
+            spectra_tpm_rf = spectra_tpm_rf.div(tpm_stats.loc[hvgs, '__std'], axis=1)
+            rf_usages = self.refit_usage(norm_tpm.X, spectra_tpm_rf)
+            rf_usages = pd.DataFrame(rf_usages, index=norm_counts.obs.index, columns=spectra_tpm_rf.index)                                                                  
+               
+        save_df_to_npz(median_spectra, self.paths['consensus_spectra']%(k, density_threshold_repl))
+        save_df_to_npz(rf_usages, self.paths['consensus_usages']%(k, density_threshold_repl))
+        save_df_to_npz(consensus_stats, self.paths['consensus_stats']%(k, density_threshold_repl))
+        save_df_to_text(median_spectra, self.paths['consensus_spectra__txt']%(k, density_threshold_repl))
+        save_df_to_text(rf_usages, self.paths['consensus_usages__txt']%(k, density_threshold_repl))
         save_df_to_npz(spectra_tpm, self.paths['gene_spectra_tpm']%(k, density_threshold_repl))
         save_df_to_text(spectra_tpm, self.paths['gene_spectra_tpm__txt']%(k, density_threshold_repl))
-
+        save_df_to_npz(usage_coef, self.paths['gene_spectra_score']%(k, density_threshold_repl))
+        save_df_to_text(usage_coef, self.paths['gene_spectra_score__txt']%(k, density_threshold_repl))
         if show_clustering:
             if topics_dist is None:
                 topics_dist = euclidean_distances(l2_spectra.values)
@@ -861,6 +937,41 @@ class cNMF():
         fig.savefig(self.paths['k_selection_plot'], dpi=250)
         if close_fig:
             plt.close(fig)
+            
+            
+    def load_results(self, K, density_threshold, n_top_genes=100):
+        """
+        Loads normalized usages and gene_spectra_scores for a given choice of K and 
+        local_density_threshold for the cNMF run. Additionally returns a DataFrame of
+        the top genes linked to each program with the number of genes indicated by the
+        `n_top_genes` parameter
+        
+        Returns
+        usage - cNMF usages (cells X K) normalized to sum to 1
+        spectra - Z-score regressed coeffecients for each program (K x genes) with higher values cooresponding
+                    to better marker genes
+        top_genes - ranked list of marker genes per GEP (n_top_genes X K)
+        """
+        scorefn = self.paths['gene_spectra_score__txt'] % (K, str(density_threshold).replace('.', '_'))
+        tpmfn = self.paths['gene_spectra_tpm__txt'] % (K, str(density_threshold).replace('.', '_'))
+        usagefn = self.paths['consensus_usages__txt'] % (K, str(density_threshold).replace('.', '_'))
+        spectra_scores = pd.read_csv(scorefn, sep='\t', index_col=0).T
+        spectra_tpm = pd.read_csv(tpmfn, sep='\t', index_col=0).T
+
+        usage = pd.read_csv(usagefn, sep='\t', index_col=0)
+        usage = usage.div(usage.sum(axis=1), axis=0)
+        
+        try:
+            usage.columns = [int(x) for x in usage.columns]
+        except:
+            print('Usage matrix columns include non integer values')
+    
+        top_genes = []
+        for gep in spectra_scores.columns:
+            top_genes.append(list(spectra_scores.sort_values(by=gep, ascending=False).index[:n_top_genes]))
+        
+        top_genes = pd.DataFrame(top_genes, index=spectra_scores.columns).T
+        return(usage, spectra_scores, spectra_tpm, top_genes)
 
 
 def main():
@@ -901,6 +1012,7 @@ def main():
     parser.add_argument('--numgenes', type=int, help='[prepare] Number of high variance genes to use for matrix factorization.', default=2000)
     parser.add_argument('--tpm', type=str, help='[prepare] Pre-computed (cell x gene) TPM values as df.npz or tab separated txt file. If not provided TPM will be calculated automatically', default=None)
     parser.add_argument('--beta-loss', type=str, choices=['frobenius', 'kullback-leibler', 'itakura-saito'], help='[prepare] Loss function for NMF.', default='frobenius')
+    parser.add_argument('--init', type=str, choices=['random', 'nndsvd'], help='[prepare] Initialization algorithm for NMF.', default='random')
     parser.add_argument('--densify', dest='densify', help='[prepare] Treat the input data as non-sparse', action='store_true', default=False) 
     parser.add_argument('--worker-index', type=int, help='[factorize] Index of current worker (the first worker should have index 0)', default=0)
     parser.add_argument('--local-density-threshold', type=float, help='[consensus] Threshold for the local density filtering. This string must convert to a float >0 and <=2', default=0.5)
@@ -914,7 +1026,7 @@ def main():
     if args.command == 'prepare':
         cnmf_obj.prepare(args.counts, components=args.components, n_iter=args.n_iter, densify=args.densify,
                          tpm_fn=args.tpm, seed=args.seed, beta_loss=args.beta_loss,
-                         num_highvar_genes=args.numgenes, genes_file=args.genes_file)
+                         num_highvar_genes=args.numgenes, genes_file=args.genes_file, init=args.init)
 
     elif args.command == 'factorize':
         cnmf_obj.factorize(worker_i=args.worker_index, total_workers=args.total_workers)
